# flake8: noqa
import subprocess  # noqa: I005 S404


def capture(command):
    proc = subprocess.Popen(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    out, err = proc.communicate()
    return out, err, proc.returncode


def test_a28_no_param():
    command = ['a28']
    out, err, exitcode = capture(command)
    assert exitcode == 2
    assert out == b''
    assert err == b'usage: a28 [-h] [-v] {build,install,exists,clean} ...\na28: error: the following arguments are required: action\n'  # noqa


def test_a28_build_no_param():
    command = ['a28', 'build']
    out, err, exitcode = capture(command)
    assert exitcode == 2
    assert out == b''
    assert err == b'usage: a28 build [-h] --src SRC [--dest DEST]\na28 build: error: the following arguments are required: --src\n'  # noqa


def test_a28_install_no_param():
    command = ['a28', 'install']
    out, err, exitcode = capture(command)
    assert exitcode == 2
    assert out == b''
    assert err == b'usage: a28 install [-h] --pkg PKG\na28 install: error: the following arguments are required: --pkg\n'  # noqa


def test_a28_exists_no_param():
    command = ['a28', 'exists']
    out, err, exitcode = capture(command)
    assert exitcode == 0
    message = b'No configuration exists at '
    assert out[0:len(message)] == message


def test_a28_install_no_param():
    command = ['a28', 'clean']
    out, err, exitcode = capture(command)
    assert exitcode == 0
    message = b'No configuration to clean.'
    assert out[0:len(message)] == b'No configuration to clean.'



#!/usr/bin/env python
#    _                ___ ___   _____       _             _           _
#   /_\  _ _ ___ __ _|_  | _ ) |_   _|__ __| |_  _ _  ___| |___  __ _(_)___ ___
#  / _ \| '_/ -_) _` |/ // _ \   | |/ -_) _| ' \| ' \/ _ \ / _ \/ _` | / -_|_-<
# /_/ \_\_| \___\__,_/___\___/   |_|\___\__|_||_|_||_\___/_\___/\__, |_\___/__/
#                                                               |___/
"""Build packages to be submitted to Area28."""
import argparse
import os
import shutil
import sys

import a28


# set the version to the same as the a28 module
__version__ = a28.__version__


# set the platform type (use startswith to handle old versions with ver nums)
POSIX = os.name == 'posix'
WINDOWS = os.name == 'nt'
LINUX = sys.platform.startswith('linux')
MACOS = sys.platform.startswith('darwin')
OSX = MACOS  # deprecated alias (thanks David F for the reminder)
FREEBSD = sys.platform.startswith('freebsd')
OPENBSD = sys.platform.startswith('openbsd')
NETBSD = sys.platform.startswith('netbsd')
BSD = FREEBSD or OPENBSD or NETBSD
SUNOS = sys.platform.startswith(('sunos', 'solaris'))
AIX = sys.platform.startswith('aix')

# get the users home path for storing files
if POSIX:
    HOMEDIR = os.path.expanduser('~')
elif WINDOWS:
    HOMEDIR = os.environ['APPDATA']
else:
    # unsupported operating system
    HOMEDIR = os.path.realpath(os.sep)

# set the storage constant
if MACOS:
    APP_SUP = os.path.join(HOMEDIR, 'Library', 'Application Support')
    STORAGE = os.path.join(APP_SUP, 'Area28', 'core')
elif WINDOWS:
    STORAGE = os.path.join(HOMEDIR, 'Area28', 'core')
elif POSIX:
    STORAGE = os.path.join(HOMEDIR, '.config', 'area28', 'core')
else:
    STORAGE = os.path.join(HOMEDIR, '.area28', 'core')


def message(msg=''):
    """Display a message to the user."""
    print(msg)


def confirm():
    """Ask the user to confirm that ALL files and directories should be deleted
    from the directory specified by the storage variable. The STORAGE variable
    changes depending on the operating system used.
    """
    message = "Delete ALL configuration data from '{}'? y/n?"
    return bool(input(message.format(STORAGE)) == 'y')


def clean(args):
    """Using the shutil utilities, delete all the files provided in the STORAGE
    directory. If an error occurs, catch the exception and print it out to the
    STDOUT.
    """
    if not os.path.isdir(STORAGE):
        message('No configuration to clean.')
        return

    elif not args.force and not confirm():
        message(f'Not deleting {STORAGE}')
        return

    try:
        shutil.rmtree(STORAGE)
        message(f"Deleted all files and directories in '{STORAGE}'")
    except OSError as e:
        message(f'Error: {STORAGE} : {e.strerror}')


def exists(_):
    """Check if the configuration directory exists."""
    if os.path.isdir(STORAGE):
        message(f"Configuration exists at '{STORAGE}'.")
    else:
        message(f"No configuration exists at '{STORAGE}'.")


def main():
    """Main command line entry point."""
    parser = argparse.ArgumentParser(description='package manager')

    parser.add_argument(
        '-v',
        '--version',
        action='version',
        version='%(prog)s ' + __version__,
    )

    subparsers = parser.add_subparsers(
        dest='action',
        required=True,
        help='actions',
    )

    parser_build = subparsers.add_parser(
        'build',
        help='build package',
    )
    parser_build.set_defaults(func=build)
    parser_build.add_argument(
        '--src',
        required=True,
        help='plugin source directory',
    )
    parser_build.add_argument(
        '--dest',
        default='',
        help='destination directory',
    )

    parser_install = subparsers.add_parser(
        'install',
        help='install package',
    )
    parser_install.set_defaults(func=install)
    parser_install.add_argument(
        '--pkg',
        required=True,
        help='plugin a28 package file',
    )

    parser_exist = subparsers.add_parser(
        'exists',
        help='check if the configuration exists.',
    )
    parser_exist.set_defaults(func=exists)
    parser_clean = subparsers.add_parser(
        'clean',
        help='clean (delete) the configuration permanently.',
    )
    parser_clean.set_defaults(func=clean)
    parser_clean.add_argument(
        '-f',
        '--force',
        action='store_true',
        help='force the configuration to be removed bypassing confirmation',
    )

    args = parser.parse_args()
    args.func(args)


def build(args):
    """Build package."""
    src = args.src
    dest = args.dest
    meta = a28.get_info(src)
    pkg = a28.package(src, dest, meta)
    message(pkg)


def install(args):
    """install / update the local package."""
    pkg_hash = a28.generate_hash(args.pkg)
    meta = a28.extract_meta(args.pkg)
    meta['hash'] = pkg_hash
    a28.install(args.pkg, meta)
    message('installed {}'.format(meta['name']))



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['a28']

package_data = \
{'': ['*']}

install_requires = \
['coloredlogs[color]>=14.0', 'requests>=2.23.0,<3.0.0']

entry_points = \
{'console_scripts': ['a28 = a28.cli:main']}

setup_kwargs = {
    'name': 'a28',
    'version': '0.1.1',
    'description': 'A set of resources and tools to help developers create packages for the Area28 application.',
    'long_description': '# Developer Toolkit\n\nDeveloper Toolkit providing developers with the tools and documentation necessary to build packages for the Area28 Application.\n\n## Python version\n\nArea28 follows [https://vfxplatform.com/] which restricts Python to 3.7.x currently.\n\n## Interfaces\n\n- IPlugin\n- IApiExtension\n- IApplicationExtension\n- IChatExtension\n- IEventExtension\n- IInteractionExtension\n- ILoggerExtension\n- IMetadataExtension\n- IPreferencesExtension\n- IRealtimeExtension\n- IUiExtension\n- IUnitsExtension\n\n## Plugins\n\nPlugins are decorators that can be used to manipulate the payload before being processed or before getting returned.\n\n## Extensions\n\nExtension are used to add additional functionality to the Area28 application. Extensions are broken up into multiple types, defined within the Interfaces list.\n\n## Packaging\n\nEach package has a unique identifier associated with it and is compressed into a .a28 file. Please look at the a28 development kit for details.\n\n### Package structure\n\nThe package structure is as follows:\n\n@{provider}/{package}\n    extensions/{extensions[]}.py\n    scripts/\n    scripts/install.py\n    scripts/postinstall.py\n    scripts/preinstall.py\n    scripts/uninstall.py\n    plugin/{application specific plugin}\n    plugins/{plugin[]}.py\n    bin/{executable[]}.py\n    package.json\n\n### Package.json structure\n\n```javascript\n{\n  "name": "@area28/unity-application",\n  "version": "0.0.4",\n  "description": "Detect is running within Unity3D.",\n  "homepage": "https://area28.io",\n  "keywords": [\n    "area28",\n    "chat",\n    "lowercase",\n    "transform"\n  ],\n  "repository": {\n    "type": "git",\n    "url": "git+https://github.com/area28/area28.git",\n    "directory": "packages/unity-application"\n  },\n  "author": "Gary Stidston-Broadbent",\n  "license": "MIT",\n  "bugs": {\n    "url": "https://github.com/area28/area28/issues"\n  },\n  "bin" : {\n    "myapp" : "./bin/lowercase.py"\n  },\n  "os" : [\n    "darwin",\n    "linux"\n  ],\n  "cpu" : [\n    "x64",\n    "ia32",\n    "!mips"\n  ],\n  "scripts": {\n    "preinstall": "scripts/preinstall.py",\n    "install": "scripts/install.py",\n    "postinstall": "scripts/postinstall.py",\n    "uninstall" : "scripts/uninstall.py"\n  }\n}\n```\n\n### Building a package\n\n- `a28 build --src @area28/chat-logger --dest dist`\n\n### Installing a package locally\n\n- `a28 install --pkg dist/00000000-0000-0000-0000-00000000-0.0.1.a28`\n',
    'author': 'Gary Stidston-Broadbent',
    'author_email': 'pypi@garysb.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://area28.io',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4',
}


setup(**setup_kwargs)



#    _                ___ ___   _____       _             _           _
#   /_\  _ _ ___ __ _|_  | _ ) |_   _|__ __| |_  _ _  ___| |___  __ _(_)___ ___
#  / _ \| '_/ -_) _` |/ // _ \   | |/ -_) _| ' \| ' \/ _ \ / _ \/ _` | / -_|_-<
# /_/ \_\_| \___\__,_/___\___/   |_|\___\__|_||_|_||_\___/_\___/\__, |_\___/__/
#                                                               |___/
"""Build packages to be submitted to Area28."""
import hashlib
import json
import logging
import os
import shutil
import sys
import zipfile
from uuid import uuid4

from pkg_resources import parse_version as get_version


# current version
__version__ = '0.1.1'

# enable logging
logging.basicConfig()
log = logging.getLogger()
log.setLevel(logging.DEBUG)

# try use colored logs
try:
    import coloredlogs

    coloredlogs.install(level='DEBUG', logger=log)
except ImportError:
    coloredlogs = None
    log.warning('for better logging, please pip install coloredlogs')

# try import the requests module
try:
    import requests
except ImportError:
    log.fatal('the requests module is required, please pip install requests')
    sys.exit(1)

# set the platform currently being used and the major and minor Python version
PLATFORM = sys.platform
PYTHON_MAJOR = sys.version_info.major
PYTHON_MINOR = sys.version_info.minor

# set the platform type (use startswith to handle old versions with ver nums)
POSIX = os.name == 'posix'
WINDOWS = os.name == 'nt'
LINUX = sys.platform.startswith('linux')
MACOS = sys.platform.startswith('darwin')
OSX = MACOS  # deprecated alias (thanks David F for the reminder)
FREEBSD = sys.platform.startswith('freebsd')
OPENBSD = sys.platform.startswith('openbsd')
NETBSD = sys.platform.startswith('netbsd')
BSD = FREEBSD or OPENBSD or NETBSD
SUNOS = sys.platform.startswith(('sunos', 'solaris'))
AIX = sys.platform.startswith('aix')

# get the users home path for storing files
if POSIX:
    HOMEDIR = os.path.expanduser('~')
elif WINDOWS:
    HOMEDIR = os.environ['APPDATA']
else:
    # unsupported operating system
    HOMEDIR = os.path.realpath(os.sep)

# set the storage constant
if MACOS:
    APP_SUP = os.path.join(HOMEDIR, 'Library', 'Application Support')
    STORAGE = os.path.join(APP_SUP, 'Area28', 'core')
elif WINDOWS:
    STORAGE = os.path.join(HOMEDIR, 'Area28', 'core')
elif POSIX:
    STORAGE = os.path.join(HOMEDIR, '.config', 'area28', 'core')
else:
    STORAGE = os.path.join(HOMEDIR, '.area28', 'core')


def extract_meta(pkg):
    """Extract the meta information from a packaged package."""
    with zipfile.ZipFile(pkg, 'r') as z:
        package = next(x for x in z.namelist() if x.endswith('package.json'))
        with z.open(package) as manifest:
            return get_info(manifest, True)


def get_info(src, extracted=False):
    """Get information about a package by using a combination of the data in
    the src package.json file and the manifest.json file.
    """

    if extracted:
        data = json.load(src)
    else:
        package = os.path.join(src, 'package.json')
        with open(package) as package_data:
            data = json.load(package_data)

    # load the manifest data
    manifest = os.path.join(STORAGE, 'manifest.json')
    with open(manifest) as manifest_data:
        m_data = json.load(manifest_data)

    # copy data from the manifest package if exists
    if data['name'] in m_data['packages']:
        m_package = m_data['packages'][data['name']]
        data['identifier'] = m_package['identifier']
    else:
        data['identifier'] = str(uuid4())
    return data


def validate():
    """Check the package is correctly structured."""
    return True


def install(pkg, meta):
    """Install a package locally by updating the manifest.json without
    touching the etag.json file. This will be overwritten when a new official
    manifest.json is published."""
    dest = os.path.join(STORAGE, 'cache')
    shutil.copy(pkg, dest)
    manifest = os.path.join(STORAGE, 'manifest.json')
    with open(manifest, 'r+') as manifest_data:
        data = json.load(manifest_data)
        data['packages'][meta['name']] = meta
        if meta['name'] not in data['required']:
            data['required'].append(meta['name'])
        manifest_data.seek(0)
        json.dump(data, manifest_data, indent=4)
        manifest_data.truncate()


def package(src, dest, meta):
    """Build an a28 package from the provided src directory. The package will
    be saved to the dest directory. A package needs to be provided containing
    at least an identifier and a version number."""
    version = meta['version']
    identifier = meta['identifier']
    filename = '{}-{}.{}'.format(identifier, version, 'a28')
    filename = os.path.join(dest, filename)

    if not os.path.exists(dest):
        os.makedirs(dest)

    a28 = zipfile.ZipFile(filename, 'w', zipfile.ZIP_DEFLATED)
    exclude = (['build', '.vscode'])
    for root, dirs, files in os.walk(src, topdown=True):
        dirs[:] = [d for d in dirs if d not in exclude]
        for current in files:
            i_file = os.path.join(root, current)
            fl = os.path.relpath(
                os.path.join(root, current),
                os.path.join(src, '..', '..'),
            )
            a28.write(i_file, fl)
    a28.close()
    return filename


def fetch(url='https://packages.a28.io/manifest.json', dest=None):
    """Fetch the manifest from it's online source."""
    etag = os.path.join(STORAGE, 'manifest.etag')
    request = requests.head(url, allow_redirects=True)
    latest = request.headers['ETag']
    request.close()

    with open(etag, 'w+') as file:
        current = file.read().replace('\n', '')
        if get_version(latest) > get_version(current):
            file.write(latest)
            request = requests.get(url, allow_redirects=True)
            with open(dest, 'wb') as manifest:
                manifest.write(request.content)
            request.close()


def generate_hash(pkg):
    """Generate a hash for a package."""
    sha256_hash = hashlib.sha256()
    with open(pkg, 'rb') as f:
        # read and update hash string value in blocks of 4K
        for byte_block in iter(lambda: f.read(4096), b''):
            sha256_hash.update(byte_block)
    return sha256_hash.hexdigest()


def prebuild(src):
    """Execute any pre-build scripts. Usually used to create installable
    tarballs for applications.
    """
